@@ -1,14 +1,15 @@
 from .lightning import LightningRpc, RpcError, Millisatoshi
-from .plugin import Plugin, monkey_patch
+from .plugin import Plugin, monkey_patch, RpcException
 
 
-__version__ = "0.8.2"
+__version__ = "0.9.3"
 
 
 __all__ = [
     "LightningRpc",
     "Plugin",
     "RpcError",
+    "RpcException",
     "Millisatoshi",
     "__version__",
     "monkey_patch"
